(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
	typeof define === 'function' && define.amd ? define(factory) :
	(global.VirtualList = factory());
}(this, (function () { 'use strict';

	function noop() {}

	function assign(tar, src) {
		for (var k in src) tar[k] = src[k];
		return tar;
	}

	function assignTrue(tar, src) {
		for (var k in src) tar[k] = 1;
		return tar;
	}

	function appendNode(node, target) {
		target.appendChild(node);
	}

	function insertNode(node, target, anchor) {
		target.insertBefore(node, anchor);
	}

	function detachNode(node) {
		node.parentNode.removeChild(node);
	}

	function createElement(name) {
		return document.createElement(name);
	}

	function addListener(node, event, handler) {
		node.addEventListener(event, handler, false);
	}

	function removeListener(node, event, handler) {
		node.removeEventListener(event, handler, false);
	}

	function setAttribute(node, attribute, value) {
		node.setAttribute(attribute, value);
	}

	function setStyle(node, key, value) {
		node.style.setProperty(key, value);
	}

	function destroyBlock(block, lookup) {
		block.u();
		block.d();
		lookup[block.key] = null;
	}

	function outroAndDestroyBlock(block, lookup) {
		block.o(function() {
			destroyBlock(block, lookup);
		});
	}

	function updateKeyedEach(old_blocks, component, changed, get_key, dynamic, ctx, list, lookup, node, has_outro, create_each_block, intro_method, next, get_context) {
		var o = old_blocks.length;
		var n = list.length;

		var i = o;
		var old_indexes = {};
		while (i--) old_indexes[old_blocks[i].key] = i;

		var new_blocks = [];
		var new_lookup = {};
		var deltas = {};

		var i = n;
		while (i--) {
			var child_ctx = get_context(ctx, list, i);
			var key = get_key(child_ctx);
			var block = lookup[key];

			if (!block) {
				block = create_each_block(component, key, child_ctx);
				block.c();
			} else if (dynamic) {
				block.p(changed, child_ctx);
			}

			new_blocks[i] = new_lookup[key] = block;

			if (key in old_indexes) deltas[key] = Math.abs(i - old_indexes[key]);
		}

		var will_move = {};
		var did_move = {};

		var destroy = has_outro ? outroAndDestroyBlock : destroyBlock;

		function insert(block) {
			block[intro_method](node, next);
			lookup[block.key] = block;
			next = block.first;
			n--;
		}

		while (o && n) {
			var new_block = new_blocks[n - 1];
			var old_block = old_blocks[o - 1];
			var new_key = new_block.key;
			var old_key = old_block.key;

			if (new_block === old_block) {
				// do nothing
				next = new_block.first;
				o--;
				n--;
			}

			else if (!new_lookup[old_key]) {
				// remove old block
				destroy(old_block, lookup);
				o--;
			}

			else if (!lookup[new_key] || will_move[new_key]) {
				insert(new_block);
			}

			else if (did_move[old_key]) {
				o--;

			} else if (deltas[new_key] > deltas[old_key]) {
				did_move[new_key] = true;
				insert(new_block);

			} else {
				will_move[old_key] = true;
				o--;
			}
		}

		while (o--) {
			var old_block = old_blocks[o];
			if (!new_lookup[old_block.key]) destroy(old_block, lookup);
		}

		while (n) insert(new_blocks[n - 1]);

		return new_blocks;
	}

	function getSpreadUpdate(levels, updates) {
		var update = {};

		var to_null_out = {};
		var accounted_for = {};

		var i = levels.length;
		while (i--) {
			var o = levels[i];
			var n = updates[i];

			if (n) {
				for (var key in o) {
					if (!(key in n)) to_null_out[key] = 1;
				}

				for (var key in n) {
					if (!accounted_for[key]) {
						update[key] = n[key];
						accounted_for[key] = 1;
					}
				}

				levels[i] = n;
			} else {
				for (var key in o) {
					accounted_for[key] = 1;
				}
			}
		}

		for (var key in to_null_out) {
			if (!(key in update)) update[key] = undefined;
		}

		return update;
	}

	function blankObject() {
		return Object.create(null);
	}

	function destroy(detach) {
		this.destroy = noop;
		this.fire('destroy');
		this.set = noop;

		if (detach !== false) this._fragment.u();
		this._fragment.d();
		this._fragment = null;
		this._state = {};
	}

	function _differs(a, b) {
		return a != a ? b == b : a !== b || ((a && typeof a === 'object') || typeof a === 'function');
	}

	function fire(eventName, data) {
		var handlers =
			eventName in this._handlers && this._handlers[eventName].slice();
		if (!handlers) return;

		for (var i = 0; i < handlers.length; i += 1) {
			var handler = handlers[i];

			if (!handler.__calling) {
				handler.__calling = true;
				handler.call(this, data);
				handler.__calling = false;
			}
		}
	}

	function get() {
		return this._state;
	}

	function init(component, options) {
		component._handlers = blankObject();
		component._bind = options._bind;

		component.options = options;
		component.root = options.root || component;
		component.store = component.root.store || options.store;
	}

	function on(eventName, handler) {
		var handlers = this._handlers[eventName] || (this._handlers[eventName] = []);
		handlers.push(handler);

		return {
			cancel: function() {
				var index = handlers.indexOf(handler);
				if (~index) handlers.splice(index, 1);
			}
		};
	}

	function set(newState) {
		this._set(assign({}, newState));
		if (this.root._lock) return;
		this.root._lock = true;
		callAll(this.root._beforecreate);
		callAll(this.root._oncreate);
		callAll(this.root._aftercreate);
		this.root._lock = false;
	}

	function _set(newState) {
		var oldState = this._state,
			changed = {},
			dirty = false;

		for (var key in newState) {
			if (this._differs(newState[key], oldState[key])) changed[key] = dirty = true;
		}
		if (!dirty) return;

		this._state = assign(assign({}, oldState), newState);
		this._recompute(changed, this._state);
		if (this._bind) this._bind(changed, this._state);

		if (this._fragment) {
			this.fire("state", { changed: changed, current: this._state, previous: oldState });
			this._fragment.p(changed, this._state);
			this.fire("update", { changed: changed, current: this._state, previous: oldState });
		}
	}

	function callAll(fns) {
		while (fns && fns.length) fns.shift()();
	}

	function _mount(target, anchor) {
		this._fragment[this._fragment.i ? 'i' : 'm'](target, anchor || null);
	}

	function _unmount() {
		if (this._fragment) this._fragment.u();
	}

	var proto = {
		destroy,
		get,
		fire,
		on,
		set,
		_recompute: noop,
		_set,
		_mount,
		_unmount,
		_differs
	};

	/* src/VirtualList.html generated by Svelte v2.4.3 */

	function visible({ items, start, end }) {
		return items.slice(start, end).map((data, i) => {
			return { index: i + start, data };
		});
	}

	function data() {
		return { top: 0, bottom: 0, start: 0, end: 0, height: '100%', _props: {} };
	}
	var methods = {
		handleScroll() {
			const { items, start, end } = this.get();
			const { offsetHeight, scrollTop } = this.refs.viewport;

			let paddingTop = 0;
			let offset = 0;
			let i = 0;

			for (let v = 0; v < this.rows.length; v += 1) {
				this.heightMap[start + v] = this.rows[v].offsetHeight;
			}

			for (; i < items.length; i += 1) {
				offset += this.heightMap[i];
				if (offset > scrollTop) break;

				paddingTop = offset;
			}

			const newStart = i++;

			for (; i < items.length; i += 1) {
				if (offset > scrollTop + offsetHeight) break;
				offset += this.heightMap[i];
			}

			const newEnd = i;

			if (newStart === start && newEnd === end) return;

			let paddingBottom = 0;
			for (; i < items.length; i += 1) paddingBottom += this.heightMap[i];

			this.set({
				top: paddingTop,
				bottom: paddingBottom,
				start: newStart,
				end: newEnd
			});
		}
	};

	function oncreate() {
		const { items, _props } = this.get();
		const { viewport, container } = this.refs;
		const viewportHeight = viewport.offsetHeight;

		const keys = Object.keys(this.options.data).filter(key => key !== 'items' && key !== 'component');
		if (keys.length) {
			const state = this.get();
			keys.forEach(key => {
				_props[key] = state[key];
			});
			this.set({ _props });

			this.on('state', ({ changed, current }) => {
				if (!keys.some(key => changed[key])) return;

				const _props = {};
				keys.forEach(key => {
					_props[key] = current[key];
				});
				this.set({ _props });
			});
		}

		this.rows = container.getElementsByClassName('row');
		this.heightMap = [];

		let height = 0;
		let i = 0;

		while (height < viewportHeight && i < items.length) {
			this.set({ end: i + 1 });

			const rowHeight = this.heightMap[i] = this.rows[i].offsetHeight;
			height += rowHeight;

			i += 1;
		}

		const end = i;
		const avg = Math.round(height / i);

		for (; i < items.length; i += 1) this.heightMap[i] = avg;

		this.set({
			bottom: (items.length - end) * avg
		});
	}
	function add_css() {
		var style = createElement("style");
		style.id = 'svelte-xo4jl9-style';
		style.textContent = "[svelte-ref-viewport].svelte-xo4jl9{overflow-y:auto;-webkit-overflow-scrolling:touch}.row.svelte-xo4jl9{overflow:hidden}";
		appendNode(style, document.head);
	}

	function create_main_fragment(component, ctx) {
		var div, div_1, each_blocks_1 = [], each_lookup = blankObject();

		var each_value = ctx.visible;

		const get_key = ctx => ctx.item.index;

		for (var i = 0; i < each_value.length; i += 1) {
			let child_ctx = get_each_context(ctx, each_value, i);
			let key = get_key(child_ctx);
			each_blocks_1[i] = each_lookup[key] = create_each_block(component, key, child_ctx);
		}

		function scroll_handler(event) {
			component.handleScroll();
		}

		return {
			c() {
				div = createElement("div");
				div_1 = createElement("div");

				for (i = 0; i < each_blocks_1.length; i += 1) each_blocks_1[i].c();
				setAttribute(div, "svelte-ref-viewport", "");
				setStyle(div_1, "padding-top", "" + ctx.top + "px");
				setStyle(div_1, "padding-bottom", "" + ctx.bottom + "px");
				addListener(div, "scroll", scroll_handler);
				setStyle(div, "height", ctx.height);
				div.className = "svelte-xo4jl9";
			},

			m(target, anchor) {
				insertNode(div, target, anchor);
				appendNode(div_1, div);

				for (i = 0; i < each_blocks_1.length; i += 1) each_blocks_1[i].m(div_1, null);

				component.refs.container = div_1;
				component.refs.viewport = div;
			},

			p(changed, ctx) {
				var each_value = ctx.visible;

				each_blocks_1 = updateKeyedEach(each_blocks_1, component, changed, get_key, 1, ctx, each_value, each_lookup, div_1, false, create_each_block, "m", null, get_each_context);

				if (changed.top) {
					setStyle(div_1, "padding-top", "" + ctx.top + "px");
				}

				if (changed.bottom) {
					setStyle(div_1, "padding-bottom", "" + ctx.bottom + "px");
				}

				if (changed.height) {
					setStyle(div, "height", ctx.height);
				}
			},

			u() {
				detachNode(div);
			},

			d() {
				for (i = 0; i < each_blocks_1.length; i += 1) each_blocks_1[i].d();

				if (component.refs.container === div_1) component.refs.container = null;
				removeListener(div, "scroll", scroll_handler);
				if (component.refs.viewport === div) component.refs.viewport = null;
			}
		};
	}

	// (3:2) {#each visible as item (item.index)}
	function create_each_block(component, key_1, ctx) {
		var div, switch_instance_spread_levels;

		var switch_value = ctx.component;

		function switch_props(ctx) {
			var switch_instance_initial_data = {};
			switch_instance_spread_levels = [
				ctx._props,
				ctx.item.data
			];

			for (var i = 0; i < switch_instance_spread_levels.length; i += 1) {
				switch_instance_initial_data = assign(switch_instance_initial_data, switch_instance_spread_levels[i]);
			}
			return {
				root: component.root,
				data: switch_instance_initial_data
			};
		}

		if (switch_value) {
			var switch_instance = new switch_value(switch_props(ctx));
		}

		return {
			key: key_1,

			first: null,

			c() {
				div = createElement("div");
				if (switch_instance) switch_instance._fragment.c();
				div.className = "row svelte-xo4jl9";
				this.first = div;
			},

			m(target, anchor) {
				insertNode(div, target, anchor);

				if (switch_instance) {
					switch_instance._mount(div, null);
				}
			},

			p(changed, ctx) {
				if (switch_value !== (switch_value = ctx.component)) {
					if (switch_instance) switch_instance.destroy();

					if (switch_value) {
						switch_instance = new switch_value(switch_props(ctx));
						switch_instance._fragment.c();
						switch_instance._mount(div, null);
					}
				}

				else {
					var switch_instance_changes = getSpreadUpdate(switch_instance_spread_levels, [
						changed._props && ctx._props,
						changed.visible && ctx.item.data
					]);
					switch_instance._set(switch_instance_changes);
				}
			},

			u() {
				detachNode(div);
			},

			d() {
				if (switch_instance) switch_instance.destroy(false);
			}
		};
	}

	function get_each_context(ctx, list, i) {
		return assign(assign({}, ctx), {
			item: list[i],
			each_value: list,
			item_index: i
		});
	}

	function VirtualList(options) {
		init(this, options);
		this.refs = {};
		this._state = assign(data(), options.data);
		this._recompute({ items: 1, start: 1, end: 1 }, this._state);

		if (!document.getElementById("svelte-xo4jl9-style")) add_css();

		if (!options.root) {
			this._oncreate = [];
			this._beforecreate = [];
			this._aftercreate = [];
		}

		this._fragment = create_main_fragment(this, this._state);

		this.root._oncreate.push(() => {
			oncreate.call(this);
			this.fire("update", { changed: assignTrue({}, this._state), current: this._state });
		});

		if (options.target) {
			this._fragment.c();
			this._mount(options.target, options.anchor);

			this._lock = true;
			callAll(this._beforecreate);
			callAll(this._oncreate);
			callAll(this._aftercreate);
			this._lock = false;
		}
	}

	assign(VirtualList.prototype, proto);
	assign(VirtualList.prototype, methods);

	VirtualList.prototype._recompute = function _recompute(changed, state) {
		if (changed.items || changed.start || changed.end) {
			if (this._differs(state.visible, (state.visible = visible(state)))) changed.visible = true;
		}
	};

	return VirtualList;

})));
